@@ -23,7 +23,7 @@
 */
 
 #include <config.h>
-#include <libgnome/gnome-i18n.h>
+#include <glib/gi18n.h>
 #include <string.h>
 #include "gtkhtml.h"
 #include "persist-stream.h"
@@ -131,7 +131,7 @@ impl_load (PortableServer_Servant servant, const Bonobo_Stream stream, const COR
 	if (was_editable)
 		gtk_html_set_editable (persist->html, FALSE);
 	
-	/* bonobo stream are _always_ utf-8 */
+	/* bonobo streams are _always_ utf-8 */
 	handle = gtk_html_begin_content (persist->html, "text/html; charset=utf-8");
 
 	do {
