@@ -375,6 +375,7 @@ merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e, GList **left, GList **
 		HTML_CLUE  (cf1)->halign = HTML_CLUE (cf2)->halign;
 		HTML_CLUE  (cf1)->valign = HTML_CLUE (cf2)->valign;
 		html_object_copy_data_from_object (self, with);
+		cf1->dir = cf2->dir;
 	}
 
 	rv = (* HTML_OBJECT_CLASS (parent_class)->merge) (self, with, e, left, right, cursor);
@@ -592,7 +593,8 @@ get_level_indent (HTMLClueFlow *flow,
 	if (flow->levels->len > 0 || ! is_item (flow)) {
 		guint cite_width, indent_width;
 
-		cite_width   = html_painter_get_block_cite_width (painter, GTK_HTML_FONT_STYLE_SIZE_3, NULL, dir);
+		cite_width = html_painter_get_block_cite_width (painter, GTK_HTML_FONT_STYLE_SIZE_3, NULL, dir)
+			+ html_painter_get_space_width (painter, GTK_HTML_FONT_STYLE_SIZE_3, NULL);
 		indent_width = html_painter_get_block_indent_width (painter, GTK_HTML_FONT_STYLE_SIZE_3, NULL);
 		
 		while (i <= level) {
@@ -686,18 +688,37 @@ calc_min_width (HTMLObject *o,
 }
 
 static gint
-pref_right_margin (HTMLPainter *p, HTMLClueFlow *clueflow, HTMLObject *o, gint y, gboolean with_aligned) 
+pref_left_margin (HTMLPainter *p, HTMLObject *o, int indent)
 {
-	gint fixed_margin = html_object_get_right_margin (o, p, y, with_aligned);
+	int margin = html_object_get_left_margin (o->parent, p, o->y, TRUE);
 
-	/* FIXME: this hack lets us wrap the display at 72 characters when we are using
-	   a plain painter */
-	  
-	if (clueflow->style == HTML_CLUEFLOW_STYLE_PRE || ! HTML_IS_PLAIN_PAINTER(p))
-		return fixed_margin;
+	if (html_object_get_direction (o) == HTML_DIRECTION_RTL) {
+		if (HTML_CLUEFLOW (o)->style != HTML_CLUEFLOW_STYLE_PRE && HTML_IS_PLAIN_PAINTER(p))
+			return MAX (margin, o->width - (int)(72 * (MAX (html_painter_get_space_width (p, GTK_HTML_FONT_STYLE_SIZE_3 | GTK_HTML_FONT_STYLE_FIXED, NULL),
+									html_painter_get_e_width (p, GTK_HTML_FONT_STYLE_SIZE_3 | GTK_HTML_FONT_STYLE_FIXED, NULL)))));
+	} else {
+		if (indent > margin)
+			margin = indent;
+	}
 
-	return MIN (fixed_margin, 72 * (MAX (html_painter_get_space_width (p, GTK_HTML_FONT_STYLE_SIZE_3 | GTK_HTML_FONT_STYLE_FIXED, NULL),
-					     html_painter_get_e_width (p, GTK_HTML_FONT_STYLE_SIZE_3 | GTK_HTML_FONT_STYLE_FIXED, NULL))));
+	return margin;
+}
+
+static gint
+pref_right_margin (HTMLPainter *p, HTMLObject *o, int indent)
+{
+	gint margin = html_object_get_right_margin (o->parent, p, o->y, TRUE);
+
+	if (html_object_get_direction (o) != HTML_DIRECTION_RTL) {
+		if (HTML_CLUEFLOW (o)->style != HTML_CLUEFLOW_STYLE_PRE && HTML_IS_PLAIN_PAINTER(p))
+			return MIN (margin, 72 * (MAX (html_painter_get_space_width (p, GTK_HTML_FONT_STYLE_SIZE_3 | GTK_HTML_FONT_STYLE_FIXED, NULL),
+						       html_painter_get_e_width (p, GTK_HTML_FONT_STYLE_SIZE_3 | GTK_HTML_FONT_STYLE_FIXED, NULL))));
+	} else {
+		if (indent > o->width - margin)
+			margin = o->width - indent;
+	}
+
+	return margin;
 }
 
 static void
@@ -721,10 +742,8 @@ add_clear_area (GList **changed_objs, HTMLObject *o, gint x, gint w)
 static void
 calc_margins (HTMLObject *o, HTMLPainter *painter, gint indent, gint *lmargin, gint *rmargin)
 {
-	*lmargin = html_object_get_left_margin (o->parent, painter, o->y, TRUE);
-	if (indent > *lmargin)
-		*lmargin = indent;
-	*rmargin = pref_right_margin (painter, HTML_CLUEFLOW (o), o->parent, o->y, TRUE);
+	*lmargin = pref_left_margin (painter, o, indent);
+	*rmargin = pref_right_margin (painter, o, indent);
 }
 
 static inline gint
@@ -1188,7 +1207,6 @@ draw_quotes (HTMLObject *self, HTMLPainter *painter,
 	gboolean is_plain = HTML_IS_PLAIN_PAINTER (painter);
 	HTMLDirection dir = html_object_get_direction (self);
 	HTMLEngine *e;
-	int space_width;
 
 	if (painter->widget && GTK_IS_HTML (painter->widget))
 		e = GTK_HTML (painter->widget)->engine;
@@ -1196,7 +1214,6 @@ draw_quotes (HTMLObject *self, HTMLPainter *painter,
 		return;
 	
 	flow = HTML_CLUEFLOW (self);
-	space_width = html_painter_get_space_width (painter, html_clueflow_get_default_font_style (flow), NULL);
 
 	for (i = 0; i < flow->levels->len; i++, last_indent = indent) {
 		indent = get_level_indent (flow, i, painter);
@@ -1239,15 +1256,13 @@ draw_quotes (HTMLObject *self, HTMLPainter *painter,
 
 
 				/* draw "> " quote characters in the plain case */ 
-				html_painter_set_font_style (painter, 
-							     html_clueflow_get_default_font_style (flow));
-
+				html_painter_set_font_style (painter, GTK_HTML_FONT_STYLE_SIZE_3 | GTK_HTML_FONT_STYLE_FIXED);
 				html_painter_set_font_face  (painter, NULL);
 
 				if (dir == HTML_DIRECTION_RTL)
-					x_pos = self->x + self->width + space_width - last_indent;
+					x_pos = self->x + self->width - last_indent;
 				else
-					x_pos = self->x + last_indent - space_width;
+					x_pos = self->x + last_indent;
 
 				draw_cite_line (self, painter, cite_str, self->ascent - baseline,
 						x_pos + tx, ty);
@@ -1591,6 +1606,28 @@ save_indent_string (HTMLClueFlow *self, HTMLEngineSaveState *state, const char *
 	return retval;
 }
 
+static char *
+get_p_str (HTMLClueFlow *self, HTMLEngineSaveState *state)
+{
+	char *p_str = NULL;
+
+	if (self->dir != html_object_get_direction (state->engine->clue)) {
+		switch (self->dir) {
+		case HTML_DIRECTION_RTL:
+			p_str = "<P DIR=RTL>\n";
+			break;
+		case HTML_DIRECTION_LTR:
+			if (html_object_get_direction (state->engine->clue) != HTML_DIRECTION_DERIVED)
+				p_str = "<P DIR=LTR>\n";
+			break;
+		default:
+			;
+		}
+	}
+
+	return p_str;
+}
+
 static gboolean
 write_flow_tag (HTMLClueFlow *self, HTMLEngineSaveState *state) 
 {
@@ -1598,13 +1635,37 @@ write_flow_tag (HTMLClueFlow *self, HTMLEngineSaveState *state)
 	HTMLClueFlow *next = NULL;
 	HTMLClueFlow *prev = NULL;
 	HTMLHAlignType halign;
-	
+	char *br_str = "<BR>\n";
+
 	if (HTML_IS_CLUEFLOW (HTML_OBJECT (self)->next))
 		next = HTML_CLUEFLOW (HTML_OBJECT (self)->next);
 	    
 	if (HTML_IS_CLUEFLOW (HTML_OBJECT (self)->prev))
 		prev = HTML_CLUEFLOW (HTML_OBJECT (self)->prev);
 
+	if (next && next->dir != html_object_get_direction (state->engine->clue)) {
+		switch (next->dir) {
+		case HTML_DIRECTION_RTL:
+			br_str = "<BR DIR=RTL>\n";
+			break;
+		case HTML_DIRECTION_LTR:
+			if (html_object_get_direction (state->engine->clue) != HTML_DIRECTION_DERIVED)
+				br_str = "<BR DIR=LTR>\n";
+			break;
+		default:
+			;
+		}
+	}
+
+	if (!prev) {
+		char *p_str = get_p_str (self, state);
+
+		if (p_str) {
+			if (! html_engine_save_output_string (state, p_str))
+				return FALSE;
+		}
+	}
+
 	d = get_similar_depth (self, prev);
 	if (is_item (self)) {
 		char *li = get_list_start_tag (self);
@@ -1652,13 +1713,13 @@ write_flow_tag (HTMLClueFlow *self, HTMLEngineSaveState *state)
 
 	if (is_item (self)) {
 		if (next && is_levels_equal (self, next) && !is_item (next) && !html_clueflow_contains_table (self)) {
-			if (!html_engine_save_output_string (state, "<BR>\n"))
+			if (!html_engine_save_output_string (state, br_str))
 				return FALSE;
 		} else if (!html_engine_save_output_string (state, "\n"))
 			return FALSE;
 	} else if (is_levels_equal (self, next) && self->style == next->style) {
 		if (self->style != HTML_CLUEFLOW_STYLE_PRE && !html_clueflow_contains_table (self)) {
-			if (!html_engine_save_output_string (state, "<BR>\n"))
+			if (!html_engine_save_output_string (state, br_str))
 				return FALSE;
 		} else {
 			if (!html_engine_save_output_string (state, "\n"))
@@ -1669,7 +1730,7 @@ write_flow_tag (HTMLClueFlow *self, HTMLEngineSaveState *state)
 
 		if (self->style != HTML_CLUEFLOW_STYLE_PRE) {
 			if ((!html_clueflow_contains_table (self) && !end && next && self->style == next->style) || html_clueflow_is_empty (self)) {
-				if (!html_engine_save_output_string (state, "<BR>\n"))
+				if (!html_engine_save_output_string (state, br_str))
 					return FALSE;
 			} else {
 				if (!html_engine_save_output_string (state, "\n"))
@@ -1686,6 +1747,21 @@ write_flow_tag (HTMLClueFlow *self, HTMLEngineSaveState *state)
 		}
 	}
 	
+	if (!next) {
+		if (HTML_OBJECT (self)->parent && html_object_is_clue (HTML_OBJECT (self)->parent)) {
+			HTMLObject *head = HTML_CLUE (HTML_OBJECT (self)->parent)->head;
+
+			if (head && HTML_IS_CLUEFLOW (head)) {
+				char *head_p_str = get_p_str (HTML_CLUEFLOW (head), state);
+
+				if (head_p_str) {
+					if (! html_engine_save_output_string (state, "</P>\n"))
+						return FALSE;
+				}
+			}
+		}
+	}
+
 	return TRUE;
 }
 
@@ -1777,7 +1853,7 @@ plain_padding (HTMLClueFlow *flow, GString *out, gboolean firstline)
 		for (i = 0; i < flow->levels->len; i++) {
 			switch (flow->levels->data[i]) {
 			case HTML_LIST_TYPE_BLOCKQUOTE_CITE:
-				APPEND_PLAIN (html_object_get_direction (HTML_OBJECT (flow)) == HTML_DIRECTION_RTL ? HTML_BLOCK_CITE_RTL : HTML_BLOCK_CITE_LTR);
+				APPEND_PLAIN (html_object_get_direction (HTML_OBJECT (flow)) == HTML_DIRECTION_RTL ? (" " HTML_BLOCK_CITE_RTL) : (HTML_BLOCK_CITE_LTR " "));
 				break;
 			case HTML_LIST_TYPE_GLOSSARY_DL:
 				break;
@@ -1862,6 +1938,7 @@ save_plain (HTMLObject *self,
 {
 	HTMLClueFlow *flow;
 	HTMLEngineSaveState *buffer_state;
+	HTMLDirection dir = html_object_get_direction (self);
 	GString *out = g_string_new ("");
 	gint pad;
 	gint align_pad;
@@ -1991,13 +2068,19 @@ save_plain (HTMLObject *self,
 
 				switch (html_clueflow_get_halignment (flow)) {
 				case HTML_HALIGN_RIGHT:
-					align_pad = max_width - width;
+					if (dir != HTML_DIRECTION_RTL)
+						align_pad = max_width - width;
+					else
+						align_pad = 0;
 					break;
 				case HTML_HALIGN_CENTER:
 					align_pad = (max_width - width) / 2;
 					break;
 				default:
-					align_pad = 0;
+					if (dir != HTML_DIRECTION_RTL)
+						align_pad = 0;
+					else
+						align_pad = max_width - width;
 					break;
 				}
 			
@@ -2365,17 +2448,7 @@ html_clueflow_real_get_direction (HTMLObject *o)
 				if (child == HTML_CLUE (o)->head && html_object_get_length (child) == 0 && o->prev)
 					return html_object_get_direction (o->prev);
 
-				PangoDirection pdir = html_text_get_pango_direction (HTML_TEXT (child));
-				switch (pdir) {
-				case PANGO_DIRECTION_RTL:
-					dir = HTML_DIRECTION_RTL;
-					break;
-				case PANGO_DIRECTION_LTR:
-					dir = HTML_DIRECTION_LTR;
-					break;
-				default:
-					break;
-				}
+				dir = html_text_direction_pango_to_html (html_text_get_pango_direction (HTML_TEXT (child)));
 
 				break;
 			}
