@@ -370,8 +370,7 @@ html_engine_new_text (HTMLEngine *e, const gchar *text, gint len)
 	if (e->insertion_font_style != GTK_HTML_FONT_STYLE_DEFAULT)
 		html_text_set_style_in_range (HTML_TEXT (to), e->insertion_font_style, e, 0, HTML_TEXT (to)->text_bytes);
 	if (e->insertion_color &&
-	    ((e->insertion_url == NULL && e->insertion_color != html_colorset_get_color (e->settings->color_set, HTMLTextColor))
-	     || (e->insertion_url && e->insertion_color != html_colorset_get_color (e->settings->color_set, HTMLLinkColor))))
+	    e->insertion_color != html_colorset_get_color (e->settings->color_set, HTMLTextColor))
 		html_text_set_color_in_range (HTML_TEXT (to), e->insertion_color, 0, HTML_TEXT (to)->text_bytes);
 	if (e->insertion_url)
 		html_text_append_link (HTML_TEXT (to), e->insertion_url, e->insertion_target, 0, HTML_TEXT (to)->text_len);
@@ -711,6 +710,7 @@ html_engine_next_cell (HTMLEngine *e, gboolean create)
 		if (create && HTML_IS_TABLE (e->cursor->object)) {
 			html_cursor_backward (e->cursor, e);
 			html_engine_insert_table_row (e, TRUE);
+			html_cursor_forward (e->cursor, e);
 		}
 		html_engine_show_cursor (e);
 
@@ -750,3 +750,15 @@ html_engine_set_title (HTMLEngine *e, const gchar *title)
 	e->title = g_string_new (title);
 	g_signal_emit_by_name (e, "title_changed");
 }
+
+void html_engine_edit_set_direction (HTMLEngine *e,
+				     HTMLDirection dir)
+{
+	HTMLClueFlow *cf = html_object_get_flow (e->cursor->object);
+
+	if (cf && cf->dir != dir && html_clueflow_is_empty (cf)) {
+		html_engine_freeze (e);
+		cf->dir = dir;
+		html_engine_thaw (e);
+	}
+}
